import {
    createSignal,
    onMount,
    onCleanup,
} from "solid-js";
import {createStore} from "solid-js/store";
import Tabbed from "./Tabbed.jsx";

import styles from "./Editor.module.css";
import styles_RightPane from "./RightPane.module.css";

import LeftPane from "./LeftPane.jsx";
import MiddlePane from "./MiddlePane.jsx";
import Properties from "./Properties.jsx";
import Controls from "./Controls.jsx";

import {World, Pane} from "../stackline-wasm/stackline_wasm.js";

// let json = await (await fetch("/prime.json")).json();

let new_world = window.localStorage.getItem("world");

export default function Editor() {
    let start_world = window.localStorage.getItem("world");

    try {
        if (start_world) {
            start_world = JSON.parse(start_world);
            start_world = World.deserialize(start_world);
        } else {
            start_world = new World();
        }
    } catch (err) {
        console.error(err);
        start_world = new World();
    }

    let [world, setWorld] = createSignal(start_world, {equals: false});
    start_world = null;

    world.init = () => setWorld((world) => {
        world.init();
        return world;
    });

    world.step = () => setWorld((world) => {
        world.step();
        return world;
    });

    let [settings, setSettings] = createStore({
        selected: null,
        time: performance.now(),
        grid: true,
        picked: null,
    });

    setWorld((world) => {
        // let tile = world.get(4, 0);
        // tile.signal = {direction: "Up", stack: []};
        // tile.state = "Active";
        // world.set(4, 0, tile);
        // world.init();

        // console.log(world.toString());

        return world;
    });

    setInterval(() => {
        setSettings("time", performance.now());
    }, 100);

    let [running, setRunning] = createSignal(null);

    function step() {
        world.init();
        world.step();
    }

    function pause() {
        if (running()) {
            clearInterval(running());
            setRunning(null);
        }
    }

    function play() {
        world.init();

        if (running()) {
            clearInterval(running());
        }

        setRunning(setInterval(() => {
            world.step();
        }, 100));
    }

    function keydown(event) {
        if (event.code === "KeyS") {
            if (running()) {
                pause();
            } else if (event.shiftKey) {
                play();
            } else if (!event.ctrlKey) {
                step();
            }
        } else if (event.code === "KeyG") {
            // TODO: immediately trigger a redraw?
            setSettings("grid", !settings.grid);
        }
    }

    onMount(() => {
        window.addEventListener("keydown", keydown);
    });

    onCleanup(() => {
        window.removeEventListener("keydown", keydown);

        if (running()) {
            clearInterval(running());
            setRunning(null);
        }
    });

    let [left_pane, presets] = LeftPane({settings, setSettings, world, setWorld});

    let last_click = 0, last_selected = null;
    const DOUBLE_CLICK_DURATION = 250;
    function click() {
        if (
            Date.now() - last_click < DOUBLE_CLICK_DURATION
            && settings.selected[0] == last_selected[0]
            && settings.selected[1] == last_selected[1]
        ) {
            if (settings.selected !== null && settings.picked !== null) {
                blit(presets[settings.picked].serialized, ...settings.selected);
            }
        }
        last_selected = [...settings.selected];
        last_click = Date.now();
    }

    function blit(serialized, x, y) {
        let pane = Pane.deserialize(serialized);

        setWorld((world) => {
            try {
                pane.blit(x, y, world);
            } finally {
                return world;
            }
        });

        pane.free();
    }

    return (
        <div class={styles.Editor}>
            {left_pane}
            <MiddlePane settings={settings} world={world} setSettings={setSettings} click={click} blit={blit} />
            <div class={styles_RightPane.RightPane}>
                <Tabbed>
                    <div title="Controls">
                        <Controls settings={settings} setSettings={setSettings} world={world} setWorld={setWorld} />
                    </div>
                    <div title="Files">Files</div>
                </Tabbed>
                <Properties settings={settings} world={world} setWorld={setWorld} setSettings={setSettings} />
            </div>
        </div>
    );
}
