import {createEffect, createSignal, createMemo, untrack} from "solid-js";
import {createStore} from "solid-js/store";

import styles from "./MiddlePane.module.css";

// TODO: use something like a Semaphore to only draw when needed

const ZOOM_STRENGTH = -0.005;

export default function MiddlePane(props) {
    let {settings, setSettings, world, blit} = props;
    let mounted = true;

    let [canvas, setCanvas] = createStore({
        width: 100,
        height: 100,
        element: null,
        ctx: null,
    });

    let [view, setView] = createStore({
        cx: 0,
        cy: 0,
        zoom: 2,
        get zoom_factor() {
            return Math.ceil(Math.pow(2, this.zoom));
        },
        get tile_size() {
            return 10 * this.zoom_factor;
        },

        from_x: 0,
        to_x: 0,
        from_y: 0,
        to_y: 0,
        get width() {
            return this.to_x - this.from_x + 1;
        },
        get height() {
            return this.to_y - this.from_y + 1;
        }
    });

    let [mouse, setMouse] = createStore({
        x: 0,
        y: 0,
        click_x: 0,
        click_y: 0,
        hovering: false,
        down: false,
        touches: [],
    });

    createEffect(() => {
        let from_y = Math.floor((-view.cy - canvas.height / 2) / view.tile_size);
        let to_y = Math.ceil((-view.cy + canvas.height / 2) / view.tile_size);
        let from_x = Math.floor((-view.cx - canvas.width / 2) / view.tile_size);
        let to_x = Math.ceil((-view.cx + canvas.width / 2) / view.tile_size);

        // effect automatically batches
        setView("from_x", from_x);
        setView("from_y", from_y);
        setView("to_x", to_x);
        setView("to_y", to_y);
    });

    let getSurface = createMemo(() => {
        let _ = settings.time;
        world().set_blink_duration(settings.time / 1000.0);

        return world().draw(-view.from_x, -view.from_y, view.width, view.height);
    }, {equals: false});

    let getPanes = createMemo(() => {
        let panes = [];

        for (let name of world().panes()) {
            let pane = world().get_pane(name);

            panes.push({
                x: pane.x,
                y: pane.y,
                width: pane.width,
                height: pane.height
            });

            pane.free();
        }

        return panes;
    });

    let getHovered = createMemo(() => {
        let x = Math.floor((mouse.x - canvas.width / 2 - view.cx) / view.tile_size);
        let y = Math.floor((mouse.y - canvas.height / 2 - view.cy) / view.tile_size);

        return [x, y];
    });

    let getSelectedHint = createMemo(() => {
        let coords = settings.selected;
        if (!coords) return null;

        let full_tile = world().get(...coords);
        if (!full_tile) return null;

        let hint = getHint(full_tile.tile);
        full_tile.free();
        return hint;
    });

    let getHoveredHint = createMemo(() => {
        let coords = getHovered();
        if (!coords) return null;

        let full_tile = world().get(...coords);
        if (!full_tile) return null;

        let hint = getHint(full_tile.tile);
        full_tile.free();
        return hint;
    });

    function getHint(tile) {
        if (!tile) return null;

        let name = Object.keys(tile)[0];
        tile = tile[name];

        if (name === "Teleporter") {
            let pane = world().get_pane(tile.coordinates[0]);
            if (!pane) return null;

            let [dx, dy] = pane.position;

            pane.free();
            return [dx + tile.coordinates[1], dy + tile.coordinates[2], 1, 1];
        }

        return null;
    }

    // Draw
    function draw() {
        if (!canvas.ctx) return;

        let surface = getSurface();
        let {width, height, ctx} = canvas;
        let {tile_size, zoom_factor, cx, cy, from_x, to_x, from_y, to_y} = view;
        let view_width = view.width;
        let view_height = view.height;
        let panes = getPanes();

        function fill_cursor(x, y, w, h) {
            let x2 = Math.round(x * tile_size + cx + width / 2) - zoom_factor * 2;
            let y2 = Math.round(y * tile_size + cy + height / 2) - zoom_factor * 2;
            let w2 = Math.round(w * tile_size) + zoom_factor * 2;
            let h2 = Math.round(h * tile_size) + zoom_factor * 2;

            ctx.fillRect(x2, y2, zoom_factor, zoom_factor * 3);
            ctx.fillRect(x2 + zoom_factor, y2, zoom_factor * 2, zoom_factor);

            ctx.fillRect(x2 + w2, y2, zoom_factor, zoom_factor * 3);
            ctx.fillRect(x2 + w2 - zoom_factor * 2, y2, zoom_factor * 2, zoom_factor);

            ctx.fillRect(x2, y2 + h2 - zoom_factor * 2, zoom_factor, zoom_factor * 3);
            ctx.fillRect(x2 + zoom_factor, y2 + h2, zoom_factor * 2, zoom_factor);

            ctx.fillRect(x2 + w2, y2 + h2 - zoom_factor * 2, zoom_factor, zoom_factor * 3);
            ctx.fillRect(x2 + w2 - zoom_factor * 2, y2 + h2, zoom_factor * 2, zoom_factor);
        }

        function stroke_rect(x, y, w, h) {
            ctx.lineWidth = zoom_factor;
            ctx.strokeRect(
                Math.round(x * tile_size + cx + width / 2) - zoom_factor * 1.5,
                Math.round(y * tile_size + cy + height / 2) - zoom_factor * 1.5,
                Math.round(w * tile_size) + zoom_factor * 2,
                Math.round(h * tile_size) + zoom_factor * 2,
            );
        }

        function fill_rect(x, y, w, h) {
            ctx.fillRect(
                Math.round(x * tile_size + cx + width / 2) - zoom_factor * 1,
                Math.round(y * tile_size + cy + height / 2) - zoom_factor * 1,
                Math.round(w * tile_size) + zoom_factor * 1,
                Math.round(h * tile_size) + zoom_factor * 1,
            );
        }

        ctx.fillStyle = "#202027";
        ctx.fillRect(0, 0, width, height);

        ctx.textAlign = "left";
        ctx.textBaseline = "top";
        ctx.font = `${tile_size}px Stackline Classic`;

        ctx.strokeStyle = "rgba(128, 128, 128, 0.5)";
        for (let pane of panes) {
            stroke_rect(
                pane.x,
                pane.y,
                pane.width,
                pane.height,
            );
        }

        for (let y = from_y; y <= to_y; y++) {
            for (let x = from_x; x <= to_x; x++) {
                let x2 = x * tile_size + cx + width / 2;
                let y2 = y * tile_size + cy + height / 2;

                let index = 3 * (view_width * (y - from_y) + x - from_x);

                let ch = String.fromCharCode(surface[index]);
                if (ch !== " ") {
                    let fg = to_color(surface[index + 1]);
                    let bg = surface[index + 2] === 0 ? null : to_color(surface[index + 2]);

                    if (bg) {
                        ctx.fillStyle = `rgba(${bg.join(",")})`;
                        ctx.fillRect(
                            Math.round(x2),
                            Math.round(y2),
                            Math.round(x2 + tile_size) - Math.round(x2),
                            Math.round(y2 + tile_size) - Math.round(y2),
                        );
                    }
                    ctx.fillStyle = `rgba(${fg.join(",")})`;
                    ctx.fillText(
                        ch,
                        Math.round(x2),
                        Math.round(y2),
                    );
                } else if (settings.grid) {
                    // Draw grid
                    let inside = false;

                    for (let pane of panes) {
                        if (x >= pane.x && x < pane.x + pane.width && y >= pane.y && y < pane.y + pane.height) {
                            inside = true;
                            break;
                        }
                    }

                    if (inside || view_width < 100 && view_height < 100) {
                        ctx.fillStyle = inside ? "#404040" : "#303030";

                        ctx.fillRect(
                            Math.round(x2 + tile_size / 2) - zoom_factor,
                            Math.round(y2 + tile_size / 2) - zoom_factor,
                            zoom_factor,
                            zoom_factor
                        );
                    }
                }
            }
        }

        if (mouse.hovering) {
            let [x, y] = getHovered();

            ctx.strokeStyle = "rgba(230, 255, 230, 0.2)";
            ctx.fillStyle = "rgba(230, 255, 230, 0.07)";
            stroke_rect(x, y, 1, 1);
            fill_rect(x, y, 1, 1);

            if (getHoveredHint() !== null) {
                ctx.fillStyle = "rgba(230, 255, 230, 0.07)";
                fill_cursor(...getHoveredHint());
            }
        }

        if (settings.selected) {
            let [x, y] = settings.selected;
            ctx.strokeStyle = "rgba(230, 230, 255, 0.1)";
            stroke_rect(x, y, 1, 1);

            if (getSelectedHint() !== null) {
                ctx.fillStyle = "rgba(230, 230, 255, 0.1)";
                fill_cursor(...getSelectedHint());
            }
        }
    }

    let running = false;
    function loop() {
        let should_redraw = true;

        function cb() {
            if (should_redraw) {
                draw();
                should_redraw = false;
            }

            if (mounted) {
                window.requestAnimationFrame(cb);
            }
        }

        if (!running) {
            running = true;

            untrack(cb);

            createEffect(() => {
                let _ = [
                    view.cx, view.cy,
                    canvas.width, canvas.height,
                    mouse.x, mouse.y,
                    getPanes(), getHovered(),
                    settings.selected, settings.time,
                ];
                should_redraw = true;
            });
        }
    }

    function resize_cb() {
        if (!canvas.element) return;
        setCanvas((obj) => {
            let dpr = window.devicePixelRatio ?? 1;

            let width = Math.round(container.clientWidth * dpr);
            let height = Math.round(container.clientHeight * dpr);

            obj.element.style.width = `${width / dpr}px`;
            obj.element.style.height = `${height / dpr}px`;

            return {
                ...obj,
                width: obj.element.width = width,
                height: obj.element.height = height,
            };
        });
    }

    let resize_listener = new ResizeObserver(resize_cb);

    function set_resize() {
        setTimeout(() => {
            resize_listener.observe(container);
        }, 0);
    }

    function remove_resize() {
        setTimeout(() => {
            resize_listener.unobserve(container);
        }, 0);
    }

    createEffect(() => {
        resize_cb();
        loop();
    });

    let _canvas, container;
    return (<div
        class={styles.MiddlePane}
        onMount={set_resize()}
        onCleanup={remove_resize()}
        ref={container}
    >
        <canvas
            ref={(el) => {
                setCanvas("element", el);
                setCanvas("ctx", el.getContext("2d"));
            }}
            class={styles.canvas}
            width={canvas.width}
            height={canvas.height}
            onMount={() => {mounted = true}}
            onCleanup={() => {mounted = false}}
            onMouseDown={(evt) => {
                let dpr = window.devicePixelRatio ?? 1;
                let mouse_x = (evt.clientX - container.offsetLeft) * dpr;
                let mouse_y = (evt.clientY - container.offsetTop) * dpr;
                setMouse((mouse) => {
                    return {
                        ...mouse,
                        x: mouse_x,
                        y: mouse_y,
                        click_x: mouse_x,
                        click_y: mouse_y,
                        hovering: true,
                        down: true,
                    }
                });
            }}
            onMouseMove={(evt) => {
                let dpr = window.devicePixelRatio ?? 1;
                setMouse("hovering", true);

                if (mouse.down) {
                    // TODO: numerically stable solution
                    setView("cx", view.cx + (evt.clientX - container.offsetLeft) * dpr - mouse.x);
                    setView("cy", view.cy + (evt.clientY - container.offsetTop) * dpr - mouse.y);
                }

                setMouse("x", (evt.clientX - container.offsetLeft) * dpr);
                setMouse("y", (evt.clientY - container.offsetTop) * dpr);
            }}
            onMouseUp={(evt) => {
                let dpr = window.devicePixelRatio ?? 1;
                setMouse("down", false);
                setMouse("x", (evt.clientX - container.offsetLeft) * dpr);
                setMouse("y", (evt.clientY - container.offsetTop) * dpr);

                let dist = Math.sqrt((mouse.x - mouse.click_x) ** 2 + (mouse.y - mouse.click_y) ** 2);
                if (dist < 10) {
                    setSettings("selected", getHovered());
                    props.click();
                }
            }}
            onMouseEnter={() => {
                setMouse("hovering", true);
            }}
            onMouseLeave={() => {
                setMouse("down", false);
                setMouse("hovering", false);
            }}
            onTouchStart={(evt) => {
                let dpr = window.devicePixelRatio ?? 1;
                setMouse("hovering", false);

                setMouse("touches", [...evt.touches].map(touch => ({
                    x: (touch.clientX - container.offsetLeft) * dpr,
                    y: (touch.clientY - container.offsetTop) * dpr,
                    identifier: touch.identifier,
                })));

                if (evt.touches.length === 1) {
                    setMouse("down", true);
                    setMouse("x", mouse.touches[0].x);
                    setMouse("y", mouse.touches[0].y);
                    setMouse("click_x", mouse.touches[0].x);
                    setMouse("click_y", mouse.touches[0].y);
                } else {
                    setMouse("down", false);
                }
            }}
            onTouchEnd={(evt) => {
                let identifiers = [...evt.changedTouches].map(t => t.identifier);
                setMouse("touches", (arr) => {
                    return arr.filter(touch => !identifiers.includes(touch.identifier));
                });

                if (evt.touches.length === 1) {
                    setMouse("down", true);
                    setMouse("x", mouse.touches[0].x);
                    setMouse("y", mouse.touches[0].y);
                    setMouse("click_x", mouse.touches[0].x);
                    setMouse("click_y", mouse.touches[0].y);
                } else {
                    setMouse("down", false);
                }
            }}
            oncapture:touchmove={(evt) => {
                evt.preventDefault();
                let dpr = window.devicePixelRatio ?? 1;
                let touches = [...evt.touches].map(touch => ({
                    x: (touch.clientX - container.offsetLeft) * dpr,
                    y: (touch.clientY - container.offsetTop) * dpr,
                    identifier: touch.identifier,
                }));

                if (evt.touches.length === 1) {
                    let mouse_x = (evt.touches[0].clientX - container.offsetLeft) * dpr;
                    let mouse_y = (evt.touches[0].clientY - container.offsetTop) * dpr;

                    setView("cx", view.cx + mouse_x - mouse.x);
                    setView("cy", view.cy + mouse_y - mouse.y);

                    setMouse("x", mouse_x);
                    setMouse("y", mouse_y);
                } else if (evt.touches.length === 2) {
                    let mx = (touches[0].x + touches[1].x) / 2;
                    let my = (touches[0].y + touches[1].y) / 2;
                    let dx = mx - (mouse.touches[0].x + mouse.touches[1].x) / 2;
                    let dy = my - (mouse.touches[0].y + mouse.touches[1].y) / 2;

                    let prev_dist = Math.sqrt(
                        (mouse.touches[0].x - mouse.touches[1].x) ** 2
                        + (mouse.touches[0].y - mouse.touches[1].y) ** 2
                    );

                    let dist = Math.sqrt(
                        (touches[0].x - touches[1].x) ** 2 + (touches[0].y - touches[1].y) ** 2
                    );

                    let old_zoom = view.zoom;
                    setView("zoom", Math.min(Math.max(view.zoom + Math.log2(dist / prev_dist), 0.0), 4.0));

                    let delta = Math.ceil(Math.pow(2, view.zoom)) / Math.ceil(Math.pow(2, old_zoom));

                    mx -= canvas.width / 2;
                    my -= canvas.height / 2;

                    setView("cx", (view.cx - mx) * delta + mx + dx);
                    setView("cy", (view.cy - my) * delta + my + dy);
                }

                setMouse("touches", touches);
            }}
            onWheel={(evt) => {
                let old_zoom = view.zoom;
                let zoom = view.zoom + evt.deltaY * ZOOM_STRENGTH;
                zoom = Math.min(Math.max(zoom, 0.0), 4.0);
                setView("zoom", zoom);

                let delta = Math.ceil(Math.pow(2, zoom)) / Math.ceil(Math.pow(2, old_zoom));
                setView("cx", view.cx * delta);
                setView("cy", view.cy * delta);
            }}
            onDrop={(evt) => {
                evt.preventDefault();

                container.classList.remove(styles.dragging);
                let data = JSON.parse(evt.dataTransfer.getData("application/json"));

                blit(data, ...getHovered());

                setSettings("selected", getHovered());
            }}
            onDragOver={(evt) => {
                let dpr = window.devicePixelRatio ?? 1;
                evt.preventDefault();

                container.classList.add(styles.dragging);
                setMouse("hovering", true);
                setMouse("x", (evt.clientX - container.offsetLeft) * dpr);
                setMouse("y", (evt.clientY - container.offsetTop) * dpr);
            }}
            onDragLeave={(evt) => {
                setMouse("hovering", false);
                container.classList.remove(styles.dragging);
            }}
        >
            Sowwy, your browser does not support canvases.
        </canvas>
    </div>);
}

export function to_color(num) {
    let alpha = (num >> 24) & 0xff;
    let red = (num >> 16) & 0xff;
    let green = (num >> 8) & 0xff;
    let blue = num & 0xff;

    return [red, green, blue, alpha];
}
