import {createEffect, createSignal, createMemo, untrack} from "solid-js";

import styles from "./RightPane.module.css";

import Signal from "./Signal.jsx";
import Direction from "./input/Direction.jsx";
import Orientation from "./input/Orientation.jsx";
import Number from "./input/Number.jsx";
import Value from "./input/Value.jsx";
import Bool from "./input/Bool.jsx";
import String from "./input/String.jsx";
import CmpOp from "./input/CmpOp.jsx";
import ArithOp from "./input/ArithOp.jsx";

export const COMPONENTS = new Map();

COMPONENTS.set("Signal", (props) => <Signal signal={props.value} setSignal={props.setValue} />);
COMPONENTS.set("Direction", Direction);
COMPONENTS.set("Orientation", Orientation);
COMPONENTS.set("Orientation|NotAny", (props) => Orientation({...props, not_any: true}));
COMPONENTS.set("Uint", (props) => <Number value={props.value} setValue={props.setValue} min={0} />);
COMPONENTS.set("Int", Number);
COMPONENTS.set("Value", Value);
COMPONENTS.set("CmpOp", CmpOp);
COMPONENTS.set("ArithOp", ArithOp);
COMPONENTS.set("Bool", Bool);
COMPONENTS.set("String", String);

export default function Tile(props) {
    let {full_tile, set_full_tile} = props;

    let tile = createMemo(() => {
        return full_tile()?.tile;
    });

    let schema = createMemo(() => {
        return full_tile()?.schema();
    });

    let tile_name = createMemo(() => {
        if (tile()) {
            return Object.keys(tile())[0];
        } else {
            return null;
        }
    });

    // This abstraction is needed to propagate the modifications of the Tile up to the FullTile
    function bindSetValue(label, value) {
        return function setValue(new_value) {
            set_full_tile((full_tile) => {
                let tile = full_tile.tile;
                let name = Object.keys(tile)[0];
                if (typeof new_value === "function") {
                    value = new_value(value);
                } else {
                    value = new_value;
                }
                tile[name] = schema_recurse(schema(), tile[name], label, value);
                full_tile.tile = tile;
                return full_tile;
            });
        }
    }

    return (<>
        <h2 class={styles.h2}><b>Tile:</b> {tile_name()}</h2>
        <Show when={tile()} fallback={<i class={styles.gray}>No tile</i>}>
            <div class={styles.indent}>
                <ol class={styles.properties}>
                    <For each={[...schema_iter(schema(), tile()[tile_name()])]} fallback={<li><i class={styles.gray}>No properties</i></li>}>
                        {([label, type, value]) => {
                            if (COMPONENTS.has(type)) {
                                let setValue = bindSetValue(label, value);
                                return (<li>
                                    {/* <label> */}
                                        <b class={styles.prop_name}>{label}: </b>
                                        {COMPONENTS.get(type)({value: () => value, setValue})}
                                    {/* </label> */}
                                </li>);
                            } else {
                                return (<li>
                                    <b class={styles.prop_name}>{label}: </b>
                                    {value?.toString()}
                                </li>);
                            }
                        }}
                    </For>
                </ol>
            </div>
        </Show>
    </>);
}

export function* schema_iter(schema, tile) {
    if (Array.isArray(schema)) {
        for (let n = 0; n < schema.length; n++) {
            yield* schema_iter(schema[n], tile[n]);
        }
    } else if (typeof schema === "object") {
        for (let name in schema) {
            yield* schema_iter(schema[name], tile[name]);
        }
    } else if (typeof schema === "string") {
        let res = schema.split(":");
        res.push(tile);
        yield res;
    }
}

export function schema_recurse(schema, tile, label, value) {
    if (Array.isArray(schema)) {
        for (let n = 0; n < schema.length; n++) {
            tile[n] = schema_recurse(schema[n], tile[n], label, value);
        }
        return tile;
    } else if (typeof schema === "object") {
        for (let name in schema) {
            tile[name] = schema_recurse(schema[name], tile[name], label, value);
        }
        return tile;
    } else if (typeof schema === "string") {
        if (schema.split(":")[0] === label) {
            return value;
        } else {
            return tile;
        }
    }
}
