use super::*;
use serde::{Deserialize, Serialize};

#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
pub enum Value {
    Number(f64),
    String(String),
}

impl Value {
    #[inline]
    pub fn as_number(&self) -> Option<f64> {
        match self {
            Value::Number(x) => Some(*x),
            _ => None,
        }
    }

    #[inline]
    pub fn as_int(&self) -> Option<i64> {
        self.as_number().map(|x| x as i64)
    }
}

impl From<f64> for Value {
    fn from(x: f64) -> Value {
        Value::Number(x)
    }
}

impl From<i32> for Value {
    fn from(x: i32) -> Value {
        Value::Number(x.into())
    }
}

impl From<String> for Value {
    fn from(string: String) -> Value {
        Value::String(string)
    }
}

impl<'a> From<&'a str> for Value {
    fn from(string: &'a str) -> Value {
        Value::String(String::from(string))
    }
}

impl std::fmt::Display for Value {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Value::Number(num) => write!(f, "{:.2}", num),
            Value::String(string) => write!(f, "\"{}\"", string),
        }
    }
}

impl std::ops::Add for Value {
    type Output = Self;

    fn add(self, other: Self) -> Self {
        use Value::*;
        use std::fmt::Write;

        match (self, other) {
            (Number(x), Number(y)) => Number(x + y),
            (String(mut x), y) => {
                write!(&mut x, "{}", y).unwrap();
                String(x)
            }
            (x, String(y)) => String(format!("{}{}", x, y)),
        }
    }
}

/// The unit of information that [`Tile`]s transmit between each other.
/// A `Signal` is made up of a [`stack`](Signal::stack), and tracks its [`position`](Signal::position) and [`direction`](Signal::direction).
///
/// ## Creating a signal
///
/// There are multiple ways to create a `Signal`:
///
/// - By cloning it, through [`clone_move`](Signal::clone_move) (recommended) or [`clone`](Signal::clone)
/// - By creating an empty signal, with [`empty`](Signal::empty)
/// - Through the [`stackline::signal!`](crate::signal!) macro
#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct Signal {
    direction: Direction,
    stack: Vec<Value>,
}

impl Signal {
    pub fn empty(direction: Direction) -> Self {
        Self {
            direction,
            stack: Vec::new(),
        }
    }

    /// Variant of [`moved`](Signal::moved), but clones the signal beforehand.
    ///
    /// See [`moved`](Signal::moved) for more information
    pub fn clone_move(&self, direction: Direction) -> Self {
        let mut res = self.clone();
        res.direction = direction;

        res
    }

    /// Sets the direction of the signal to `direction`, and returns that signal.
    ///
    /// This function or its sister function, [`clone_move`](Signal::clone_move), should always be called before [`send`ing](UpdateContext::send) a signal to another tile.
    ///
    /// ## Example
    ///
    /// ```
    /// # use stackline::tile::prelude::*;
    /// # #[derive(Clone, Debug, Serialize, Deserialize)]
    /// # struct MyTile;
    /// # impl Tile for MyTile {
    /// fn update<'b>(&'b mut self, mut ctx: UpdateContext<'b>) {
    ///     let direction = Direction::Down;
    ///
    ///     if let Some(signal) = ctx.take_signal() {
    ///         // We have a signal, see if it can be sent down
    ///         if let Some(pos) = ctx.accepts_direction(direction) {
    ///             ctx.send(pos, direction, signal);
    ///         }
    ///     }
    /// }
    /// # }
    /// ```
    pub fn moved(mut self, direction: Direction) -> Self {
        self.direction = direction;

        self
    }

    pub fn direction(&self) -> Direction {
        self.direction
    }

    pub fn set_direction(&mut self, direction: Direction) {
        self.direction = direction;
    }

    /// Pushes a value onto the stack of the signal.
    /// Signals are pushed on top of the stack and can be [`pop`ped](Signal::pop) in reverse order.
    ///
    /// ## Example
    ///
    /// ```
    /// # use stackline::prelude::*;
    /// let mut signal = Signal::empty(Direction::Down);
    /// assert_eq!(signal.len(), 0);
    ///
    /// signal.push(Value::Number(1.0));
    /// assert_eq!(signal.len(), 1);
    ///
    /// signal.push(Value::Number(2.0));
    /// assert_eq!(signal.len(), 2);
    /// ```
    pub fn push(&mut self, value: Value) {
        self.stack.push(value);
    }

    /// Pops a value from the stack of the signal, returning `Some(value)`
    /// if the stack wasn't empty and `None` otherwise.
    ///
    /// ## Example
    ///
    /// ```
    /// # use stackline::prelude::*;
    /// let mut signal = Signal::empty(Direction::Down);
    ///
    /// signal.push(Value::Number(1.0));
    /// signal.push(Value::Number(2.0));
    /// assert_eq!(signal.len(), 2);
    ///
    /// // We pushed 2.0 last, so pop() will return that
    /// assert_eq!(signal.pop(), Some(Value::Number(2.0)));
    /// assert_eq!(signal.len(), 1);
    /// ```
    pub fn pop(&mut self) -> Option<Value> {
        self.stack.pop()
    }

    /// Returns the number of elements in the stack of the signal.
    ///
    /// ## Example
    ///
    /// ```
    /// # use stackline::prelude::*;
    /// let signal = stackline::signal!(Direction::Down, [1.0, -1.5]);
    ///
    /// assert_eq!(signal.len(), 2);
    /// ```
    #[allow(clippy::len_without_is_empty)] // The signal will hold a `heap` in the future, making `is_empty` ambiguous
    pub fn len(&self) -> usize {
        self.stack.len()
    }

    /// Puts `other` on top of `self`
    pub fn append(&mut self, other: &Signal) {
        self.stack.extend(other.stack.iter().cloned());
    }

    pub fn is_stack_empty(&self) -> bool {
        self.stack.is_empty()
    }

    pub fn stack(&self) -> &Vec<Value> {
        &self.stack
    }

    pub fn stack_mut(&mut self) -> &mut Vec<Value> {
        &mut self.stack
    }
}

impl PartialEq<Signal> for Signal {
    fn eq(&self, other: &Self) -> bool {
        self.stack == other.stack
    }
}

impl Default for Signal {
    fn default() -> Self {
        Self::empty(Direction::default())
    }
}

/// Creates a signal with initial values in its stack.
///
/// The syntax for the macro is `signal!(position, direction, [value1, value2, ...])`, where:
/// - `position`: a pair `(x, y): (usize, usize)`
/// - `direction`: a [`Direction`] *(optional, defaults to [`Direction::default()`])*
/// - `value1`, `value2`, etc.: [`Value`] must implement [`From<T>`](Value#trait-implementations) for each value *(optional)*
///
/// ## Examples
///
/// ```
/// # use stackline::prelude::*;
/// // Creates an empty signal at (1, 2)
/// let signal = stackline::signal!();
///
/// // Creates an empty signal going right at (2, 2)
/// let signal = stackline::signal!(Direction::Right);
///
/// // Creates a signal with the values 10 and 12 on its stack
/// let signal = stackline::signal!([10, 12]);
/// ```
#[macro_export]
macro_rules! signal {
    ( $dir:expr, [ $( $x:expr ),* ] ) => {{
        let mut res = Signal::empty($dir);

        $({
            res.push(Value::from($x));
        })*

        res
    }};

    ( [ $( $x:expr ),* ] ) => {{
        let mut res = Signal::empty(Direction::default());

        $({
            res.push(Value::from($x));
        })*

        res
    }};

    ( $dir:expr ) => {{
        let mut res = Signal::empty($dir);

        res
    }};

    () => {{
        let mut res = Signal::empty(Direction::default());

        res
    }};
}
